/*
 * Copyright (c) 2017, Yuichiro MORIGUCHI
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * * Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Yuichiro MORIGUCHI nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Yuichiro MORIGUCHI BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Runtime.Serialization;

namespace Morilib.Sed
{
	public class Util
	{
		public const int LastAddress = 1000000;
	}

	public class SedException : Exception
	{
		public SedException() : base()
		{
		}
		public SedException(string message) : base(message)
		{
		}
		public SedException(string message, Exception inner) : base(message, inner)
		{
		}
		protected SedException(SerializationInfo info, StreamingContext context)
        {
        }
	}

	public class TextInfo
	{
		public const int LineFold = 72;
		private TextReader read;
		private TextWriter write;
		private bool useend;
		public string PatternSpace { get; set; }
		public string HoldSpace { get; set; }
		private string nextline;
		private int lineno = 1;
		public bool Deleted { get; private set; }
		public bool Succeed { get; set; }
		public bool EchoBack { get; private set; }
		private Dictionary<string, int> labelSpace;
		private Dictionary<string, StreamWriter> writers;
		public TextInfo(TextWriter write, bool useend, bool echoback,
			Dictionary<string, int> labelSpace)
		{
			this.read = null;
			this.write = write;
			EchoBack = echoback;
			this.labelSpace = new Dictionary<string, int>(labelSpace);
			writers = new Dictionary<string, StreamWriter>();
			if(useend && PatternSpace != null)
			{
				nextline = read.ReadLine();
			}
		}
		public void SetReadFile(string filename)
		{
			if(read != null)
			{
				read.Close();
			}
			read = new StreamReader(filename);
		}
		private string ReadReadLine()
		{
			if(read == null)
			{
				return Console.In.ReadLine();
			}
			else
			{
				return read.ReadLine();
			}
		}
		public void ReadFirst()
		{
			HoldSpace = "";
			lineno = 1;
			Deleted = false;
			Succeed = false;
			PatternSpace = ReadReadLine();
		}
		public void ReadNext()
		{
			Deleted = false;
			Succeed = false;
			if(useend)
			{
				PatternSpace = nextline;
				if(PatternSpace != null)
				{
					nextline = ReadReadLine();
				}
			}
			else
			{
				PatternSpace = ReadReadLine();
			}
			lineno++;
		}
		public int GetLineNumber()
		{
			return lineno;
		}
		public bool LastLine()
		{
			return nextline == null;
		}
		public void WriteLine(string str)
		{
			write.WriteLine(str);
		}
		private bool IsCtrl(int ch)
		{
			return (ch >= 0 && ch <= 0x1f) || ch == 0x7f || (ch >= 0x80 && ch <= 0x9f);
		}
		private readonly Regex FWPat = new Regex(
			@"[\p{IsCJKUnifiedIdeographs}" +
			@"\p{IsCJKCompatibilityIdeographs}" +
			@"\p{IsCJKUnifiedIdeographsExtensionA}" +
			@"\p{IsHiragana}\p{IsKatakana}\p{IsHangulSyllables}" +
			@"\uFF10-\uFF5A\u3000]");
		private int DisplayLength(int ch)
		{
			return FWPat.IsMatch("" + (char)ch) ? 2 : 1;
		}
		private string ListChar(int ch)
		{
			string st;

			switch(ch)
			{
			case '\a':
				return "\\a";
			case '\b':
				return "\\b";
			case '\f':
				return "\\f";
			case '\n':
				return "\\n";
			case '\r':
				return "\\r";
			case '\t':
				return "\\t";
			case '\v':
				return "\\v";
			default:
				if(IsCtrl(ch))
				{
					st = "00" + Convert.ToString(ch, 8);
					st = st.Substring(st.Length - 3, 3);
					return "\\" + st;
				}
				else
				{
					return "" + (char)ch;
				}
			}
		}
		public void WriteListLine(string str)
		{
			int no = 0, ln;
			string st;

			for(int i = 0; i < str.Length; i++)
			{
				st = ListChar(str[i]);
				ln = st.Length > 1 ? st.Length : DisplayLength(str[i]);
				if(no + ln >= LineFold)
				{
					write.WriteLine("\\");
					no = 0;
				}
				write.Write(st);
				no += ln;
			}
			write.WriteLine("$");
		}
		public void WriteLineToFile(string str, string filename)
		{
			StreamWriter ts;

			if(writers.ContainsKey(filename))
			{
				ts = writers[filename];
			}
			else
			{
				ts = new StreamWriter(filename);
				writers[filename] = ts;
			}
			ts.WriteLine(str);
		}
		public void Dispose()
		{
			if(read != null)
			{
				read.Close();
			}
			foreach(StreamWriter w in writers.Values)
			{
				w.Close();
			}
		}
		public string PatternSpaceHead()
		{
			int ind;

			ind = PatternSpace.IndexOf('\n');
			if(ind < 0)
			{
				return PatternSpace;
			}
			else
			{
				return PatternSpace.Substring(0, ind);
			}
		}
		public void Delete()
		{
			Deleted = true;
		}
		public void Delete1Line()
		{
			int ind;

			ind = PatternSpace.IndexOf('\n');
			if(ind < 0)
			{
				Deleted = true;
			}
			else
			{
				PatternSpace = PatternSpace.Substring(ind + 1, PatternSpace.Length - ind - 1);
			}
		}
		public int GetAddress(string label)
		{
			if(!labelSpace.ContainsKey(label))
			{
				throw new SedException("undefined label " + label);
			}
			return labelSpace[label];
		}
	}

	public interface Line
	{
		bool Match(TextInfo line);
		bool LastMatch(TextInfo line);
		int Arity();
	}

	public class LineAll : Line
	{
		public bool Match(TextInfo line)
		{
			return true;
		}
		public bool LastMatch(TextInfo line)
		{
			return true;
		}
		public int Arity()
		{
			return 0;
		}
	}

	public class LineRange : Line
	{
		private bool matching = false;
		private bool nextmatch = false;
		private bool lastmatch = false;
		private Line linebegin, lineend;
		private int lineno = -1;
		public LineRange(Line b, Line e)
		{
			linebegin = b;
			lineend = e;
		}
		private void ReadLine(TextInfo line)
		{
			matching = nextmatch;
			if(matching)
			{
				nextmatch = !lineend.Match(line);
				lastmatch = lineend.Match(line);
			}
			else
			{
				matching = linebegin.Match(line);
				nextmatch = matching && !lineend.Match(line);
				lastmatch = matching && lineend.Match(line);
			}
		}
		public bool Match(TextInfo line)
		{
			if(lineno != line.GetLineNumber())
			{
				ReadLine(line);
				lineno = line.GetLineNumber();
			}
			return matching;
		}
		public bool LastMatch(TextInfo line)
		{
			if(lineno != line.GetLineNumber())
			{
				ReadLine(line);
				lineno = line.GetLineNumber();
			}
			return lastmatch;
		}
		public int Arity()
		{
			return 2;
		}
	}

	public class LineNum : Line
	{
		private int linenum;
		public LineNum(int no)
		{
			linenum = no;
		}
		public bool Match(TextInfo line)
		{
			return line.GetLineNumber() == linenum;
		}
		public bool LastMatch(TextInfo line)
		{
			return Match(line);
		}
		public int Arity()
		{
			return 1;
		}
	}

	public class LineRE : Line
	{
		private Sedre re;
		public LineRE(string re)
		{
			this.re = Sedre.Pattern(re);
		}
		public bool Match(TextInfo line)
		{
			return re.Match(line.PatternSpace);
		}
		public bool LastMatch(TextInfo line)
		{
			return Match(line);
		}
		public int Arity()
		{
			return 1;
		}
	}

	public class LineLast : Line
	{
		public bool Match(TextInfo line)
		{
			return line.LastLine();
		}
		public bool LastMatch(TextInfo line)
		{
			return Match(line);
		}
		public int Arity()
		{
			return 1;
		}
	}

	public class LineNegate : Line
	{
		private Line wrapee;
		public LineNegate(Line wrapee)
		{
			this.wrapee = wrapee;
		}
		public bool Match(TextInfo line)
		{
			return !wrapee.Match(line);
		}
		public bool LastMatch(TextInfo line)
		{
			return false;
		}
		public int Arity()
		{
			return wrapee.Arity();
		}
	}

	public interface CommandLine {
		int Execute(int addr, TextInfo info);
		bool InAddress(int addr);
		int PreExecute(int addr, TextInfo info);
		int PostExecute(int addr, TextInfo info);
	}

	public abstract class SimpleCommandLine : CommandLine
	{
		protected Line line;
		protected int address;
		protected SimpleCommandLine(int addr, Line line)
		{
			this.line = line;
			this.address = addr;
		}
		public abstract int ExecuteCommand(int addr, TextInfo info);
		public virtual int PreExecuteCommand(int addr, TextInfo info)
		{
			return addr + 1;
		}
		public virtual int PostExecuteCommand(int addr, TextInfo info)
		{
			return addr + 1;
		}
		public int Execute(int addr, TextInfo info)
		{
			if(line.Match(info))
			{
				return ExecuteCommand(addr, info);
			}
			else
			{
				return addr + 1;
			}
		}
		public int PreExecute(int addr, TextInfo info)
		{
			if(line.Match(info))
			{
				return PreExecuteCommand(addr, info);
			}
			else
			{
				return addr + 1;
			}
		}
		public int PostExecute(int addr, TextInfo info)
		{
			if(line.Match(info))
			{
				return PostExecuteCommand(addr, info);
			}
			else
			{
				return addr + 1;
			}
		}
		public bool InAddress(int addr)
		{
			return addr == address;
		}
	}

	public class Commandp : SimpleCommandLine
	{
		private bool head;
		public Commandp(int addr, Line line, bool head) : base(addr, line)
		{
			this.head = head;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			info.WriteLine(head ? info.PatternSpaceHead() : info.PatternSpace);
			return addr + 1;
		}
	}

	public class Commandq : SimpleCommandLine
	{
		public Commandq(int addr, Line line) : base(addr, line)
		{
			if(line.Arity() > 1)
			{
				throw new SedException("invalid line selecion");
			}
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			return -1;
		}
	}

	public class Commandd : SimpleCommandLine
	{
		private bool head;
		public Commandd(int addr, Line line, bool head) : base(addr, line)
		{
			this.head = head;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			if(head)
			{
				info.Delete1Line();
				return 0;
			}
			else
			{
				info.Delete();
				return Util.LastAddress;
			}
		}
	}

	public class Commanda : SimpleCommandLine
	{
		private string text;
		public Commanda(int addr, Line line, string text) : base(addr, line)
		{
			if(line.Arity() > 1)
			{
				throw new SedException("invalid line selecion");
			}
			this.text = text;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			return addr + 1;
		}
		public override int PostExecuteCommand(int addr, TextInfo info)
		{
			info.WriteLine(text);
			return addr + 1;
		}
	}

	public class Commandi : SimpleCommandLine
	{
		private string text;
		public Commandi(int addr, Line line, string text) : base(addr, line)
		{
			if(line.Arity() > 1)
			{
				throw new SedException("invalid line selecion");
			}
			this.text = text;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			return addr + 1;
		}
		public override int PreExecuteCommand(int addr, TextInfo info)
		{
			info.WriteLine(text);
			return addr + 1;
		}
	}

	public class Commandc : SimpleCommandLine
	{
		private string text;
		public Commandc(int addr, Line line, string text) : base(addr, line)
		{
			this.text = text;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			info.Delete();
			return Util.LastAddress;
		}
		public override int PostExecuteCommand(int addr, TextInfo info)
		{
			if(line.LastMatch(info))
			{
				info.WriteLine(text);
			}
			return addr + 1;
		}
	}

	public class SubstFlags
	{
		public bool Print { get; set; }
		public int Occur { get; set; }
		public SubstFlags()
		{
			Print = false;
			Occur = 1;
		}
	}

	public class Commands : SimpleCommandLine
	{
		private Sedre regex;
		private string substitute;
		private SubstFlags flags;
		public Commands(int addr, Line line, string re, string repl, SubstFlags flags) : base(addr, line)
		{
			regex = Sedre.Pattern(re);
			substitute = repl;
			this.flags = flags;
		}
	    private string SubstituteString0(string line, int mbegin, int mend)
    	{
	        StringBuilder buf = new StringBuilder();
	        string group;
    	    int ch;

    	    for(int i = 0; i < substitute.Length; i++)
        	{
            	ch = substitute[i];
	            if(ch == '\\')
    	        {
        	        i++;
            	    if(i >= substitute.Length)
                	{
                    	buf.Append('\\');
	                }
    	            else if(substitute[i] >= '0' && substitute[i] <= '9')
        	        {
            	        group = regex.GetCapture(substitute[i] - '0');
                	    if(group != null)
                    	{
                        	buf.Append(group);
	                    }
    	            }
        	        else
            	    {
                	    buf.Append(substitute[i]);
	                }
    	        }
        	    else if(ch == '&')
            	{
                	buf.Append(regex.GetCapture(0));
	            }
    	        else
        	    {
            	    buf.Append((char)ch);
	            }
    	    }
	        return buf.ToString();
	    }
	    private string SubstituteString(string line, int mbegin, int mend)
    	{
	        StringBuilder buf = new StringBuilder();
    	    string pre = line.Substring(0, mbegin);
        	string post = line.Substring(mend, line.Length - mend);

	        buf.Append(pre);
	        buf.Append(SubstituteString0(line, mbegin, mend));
        	buf.Append(post);
	        return buf.ToString();
	    }
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			StringBuilder buf;
			string line = info.PatternSpace;
			bool matched = false;
			int iend = 0;

			if(flags.Occur < 0)
			{
				buf = new StringBuilder();
				if(regex.Match(line))
				{
					matched = true;
					buf.Append(line.Substring(iend, regex.MatchBegin - iend));
					buf.Append(SubstituteString0(line, regex.MatchBegin, regex.MatchEnd));
					iend = regex.MatchEnd;
					while(regex.MatchNext())
					{
						if(regex.MatchBegin == regex.MatchEnd)
						{
							throw new SedException("endless substitution");
						}
						buf.Append(line.Substring(iend, regex.MatchBegin - iend));
						buf.Append(SubstituteString0(line, regex.MatchBegin, regex.MatchEnd));
						iend = regex.MatchEnd;
					}
					buf.Append(line.Substring(iend, line.Length - iend));
					line = buf.ToString();
				}
			}
			else if(flags.Occur > 1)
			{
				if(!regex.Match(line))
				{
					// do nothing
				}
				else
				{
					for(int i = 2; regex.MatchNext(); i++)
					{
						if(flags.Occur == i)
						{
							matched = true;
							line = SubstituteString(line, regex.MatchBegin, regex.MatchEnd);
							break;
						}
					}
				}
			}
			else
			{
				if(regex.Match(line))
				{
					matched = true;
					line = SubstituteString(line, regex.MatchBegin, regex.MatchEnd);
				}
			}
			if(matched)
			{
				if(flags.Print)
				{
					info.WriteLine(line);
				}
				info.PatternSpace = line;
			}
			info.Succeed = matched;
			return addr + 1;
		}
	}

	public class Commandn : SimpleCommandLine
	{
		private bool append;
		public Commandn(int addr, Line line, bool append) : base(addr, line)
		{
			this.append = append;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			string tmp;

			if(append)
			{
				tmp = info.PatternSpace;
				info.ReadNext();
				if(info.PatternSpace == null)
				{
					return -1;
				}
				info.PatternSpace = tmp + "\n" + info.PatternSpace;
			}
			else
			{
				if(info.EchoBack)
				{
					info.WriteLine(info.PatternSpace);
				}
				info.ReadNext();
				if(info.PatternSpace == null)
				{
					return -1;
				}
			}
			return addr + 1;
		}
	}

	public class CommandJump : SimpleCommandLine
	{
		private string label;
		private bool test;
		public CommandJump(int addr, Line line, string label, bool test) : base(addr, line)
		{
			this.label = label;
			this.test = test;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			if(!test || info.Succeed)
			{
				return label != null ? info.GetAddress(label) : Util.LastAddress;
			}
			else
			{
				return addr + 1;
			}
		}
	}

	public class CommandPutNo : SimpleCommandLine
	{
		public CommandPutNo(int addr, Line line) : base(addr, line)
		{
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			info.WriteLine(info.GetLineNumber().ToString());
			return addr + 1;
		}
	}

	public class Commandh : SimpleCommandLine
	{
		private bool append;
		public Commandh(int addr, Line line, bool append) : base(addr, line)
		{
			this.append = append;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			if(append)
			{
				info.HoldSpace += "\n" + info.PatternSpace;
			}
			else
			{
				info.HoldSpace = info.PatternSpace;
			}
			return addr + 1;
		}
	}

	public class Commandg : SimpleCommandLine
	{
		private bool append;
		public Commandg(int addr, Line line, bool append) : base(addr, line)
		{
			this.append = append;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			if(append)
			{
				info.PatternSpace += "\n" + info.HoldSpace;
			}
			else
			{
				info.PatternSpace = info.HoldSpace;
			}
			return addr + 1;
		}
	}

	public class Commandx : SimpleCommandLine
	{
		public Commandx(int addr, Line line) : base(addr, line)
		{
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			string tmp;

			tmp = info.PatternSpace;
			info.PatternSpace = info.HoldSpace;
			info.HoldSpace = tmp;
			return addr + 1;
		}
	}

	public class Commandy : SimpleCommandLine
	{
		private string trbef;
		private string traft;
		public Commandy(int addr, Line line, string before, string after) : base(addr, line)
		{
			trbef = before;
			traft = after;
			if(trbef.Length != traft.Length)
			{
				throw new SedException("transform strings are not the same length");
			}
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			StringBuilder buf = new StringBuilder();
			int idx;

			for(int i = 0; i < info.PatternSpace.Length; i++)
			{
				idx = trbef.IndexOf(info.PatternSpace[i]);
				if(idx >= 0)
				{
					buf.Append(traft[idx]);
				}
				else
				{
					buf.Append(info.PatternSpace[i]);
				}
			}
			info.PatternSpace = buf.ToString();
			return addr + 1;
		}
	}

	public class Commandr : SimpleCommandLine
	{
		private string filename;
		public Commandr(int addr, Line line, string filename) : base(addr, line)
		{
			if(line.Arity() > 1)
			{
				throw new SedException("invalid line selecion");
			}
			this.filename = filename;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			return addr + 1;
		}
		public override int PostExecuteCommand(int addr, TextInfo info)
		{
			string line;

			using(StreamReader sr = new StreamReader(filename))
			{
				while((line = sr.ReadLine()) != null)
				{
					info.WriteLine(line);
				}
			}
			return addr + 1;
		}
	}

	public class Commandw : SimpleCommandLine
	{
		private string filename;
		public Commandw(int addr, Line line, string filename) : base(addr, line)
		{
			this.filename = filename;
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			info.WriteLineToFile(info.PatternSpace, filename);
			return addr + 1;
		}
	}

	public class Commandl : SimpleCommandLine
	{
		public Commandl(int addr, Line line) : base(addr, line)
		{
		}
		public override int ExecuteCommand(int addr, TextInfo info)
		{
			info.WriteListLine(info.PatternSpace);
			return addr + 1;
		}
	}

	public class CompoundCommandLine : CommandLine
	{
		private Line line;
		private List<CommandLine> commandList;
		private int addressBegin;
		private int addressEnd;
		public CompoundCommandLine(int addrbegin, int addrend, Line line, List<CommandLine> list)
		{
			addressBegin = addrbegin;
			addressEnd = addrend;
			this.line = line;
			commandList = new List<CommandLine>(list);
		}
		public int Execute(int addr, TextInfo info)
		{
			if(addr == addressBegin && !line.Match(info))
			{
				return addressEnd;
			}
			else
			{
				foreach(CommandLine c in commandList)
				{
					if(c.InAddress(addr))
					{
						return c.Execute(addr, info);
					}
				}
				throw new Exception();
			}
		}
		public bool InAddress(int addr)
		{
			return addr >= addressBegin && addr < addressEnd;
		}
		public int PreExecute(int addr, TextInfo info)
		{
			if(!line.Match(info))
			{
				return addressEnd;
			}
			else
			{
				foreach(CommandLine c in commandList)
				{
					if(c.InAddress(addr))
					{
						return c.PreExecute(addr, info);
					}
				}
				throw new Exception();
			}
		}
		public int PostExecute(int addr, TextInfo info)
		{
			if(!line.Match(info))
			{
				return addressEnd;
			}
			else
			{
				foreach(CommandLine c in commandList)
				{
					if(c.InAddress(addr))
					{
						return c.PostExecute(addr, info);
					}
				}
				throw new Exception();
			}
		}
	}

	public class SedExecuter
	{
		private CommandLine command;
		public SedExecuter(CommandLine cmd)
		{
			command = cmd;
		}
		public bool Execute(TextInfo info)
		{
			int pc = 0;

			while(command.InAddress(pc))
			{
				pc = command.Execute(pc, info);
			}
			return pc >= 0;
		}
		public void PreExecute(TextInfo info)
		{
			int pc = 0;

			while(command.InAddress(pc))
			{
				pc = command.PreExecute(pc, info);
			}
		}
		public void PostExecute(TextInfo info)
		{
			int pc = 0;

			while(command.InAddress(pc))
			{
				pc = command.PostExecute(pc, info);
			}
		}
	}

	public class SedMain
	{
		private TextInfo stream;
		private SedExecuter executer;
		public SedMain(TextWriter write, SedExecuter exec, bool echo, bool useend,
			Dictionary<string, int> labelSpace)
		{
			executer = exec;
			stream = new TextInfo(write, useend, echo, labelSpace);
		}
		public void Execute()
		{
			stream.ReadFirst();
			for(; stream.PatternSpace != null; stream.ReadNext())
			{
				executer.PreExecute(stream);
				if(!executer.Execute(stream))
				{
					break;
				}
				if(stream.EchoBack && !stream.Deleted)
				{
					stream.WriteLine(stream.PatternSpace);
				}
				executer.PostExecute(stream);
			}
		}
		public void Execute(string filename)
		{
			stream.SetReadFile(filename);
			Execute();
		}
		public void Dispose()
		{
			stream.Dispose();
		}
	}

	public class Option
	{
		public bool Echo { get; private set; }
		public string Script { get; private set; }
		private void CheckNextArg(string[] args, int acnt, string msg)
		{
			if(acnt >= args.Length)
			{
				throw new SedException(msg);
			}
		}
		private void AppendScript(string scr)
		{
			if(Script == null)
			{
				Script = scr;
			}
			else
			{
				Script = Script + "\n" + scr;
			}
		}
		public int Parse(string[] args)
		{
			int acnt = 0;
			string scr;

			Echo = true;
			Script = null;
			while(acnt < args.Length)
			{
				if(args[acnt].Equals("-n"))
				{
					Echo = false;
					acnt++;
				}
				else if(args[acnt].Equals("-f"))
				{
					acnt++;
					CheckNextArg(args, acnt, "bad argument: -f");
					using(StreamReader sr = new StreamReader(args[acnt++]))
					{
						scr = sr.ReadToEnd();
					}
					if(scr.StartsWith("#n"))
					{
						Echo = false;
					}
					AppendScript(scr);
				}
				else if(args[acnt].Equals("-e"))
				{
					acnt++;
					CheckNextArg(args, acnt, "bad argument: -e");
					AppendScript(args[acnt++]);
				}
				else
				{
					break;
				}
			}
			if(Script == null)
			{
				CheckNextArg(args, acnt, "no script");
				Script = args[acnt++];
			}
			return acnt;
		}
	}

}